#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
import requests

LINEAGE_BUILD_TARGETS_FILE = "https://raw.githubusercontent.com/LineageOS/hudson/master/lineage-build-targets"

class LineageBuildTarget:
    def __init__(self,
                 device: str,
                 build_type: str,
                 branch_name: str,
                 period: str,
                ):
        self.device = device
        self.build_type = build_type
        self.branch_name = branch_name
        self.period = period

    def __str__(self) -> str:
        return f"{self.device} {self.build_type} {self.branch_name} {self.period}"

    @classmethod
    def from_api(self, line: str):
        return self(*line.split())

def get_lineage_build_targets():
    response = requests.get(url=LINEAGE_BUILD_TARGETS_FILE).text.split("\n")
    return [LineageBuildTarget.from_api(line)
            for line in response
            if line and not line.startswith("#")]



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict


class Account:
	def __init__(self,
	             token: str,
				 email: str,
				 tier: str,
				 root_folder: str,
				 files_count: int,
				 files_count_limit: int,
				 total_size: int,
				 total_size_limit: int,
				 total_30ddl_traffic: int,
				 total_30ddl_traffic_limit: int,
	            ):
		self.token = token
		self.email = email
		self.tier = tier
		self.root_folder = root_folder
		self.files_count = files_count
		self.files_count_limit = files_count_limit
		self.total_size = total_size
		self.total_size_limit = total_size_limit
		self.total_30ddl_traffic = total_30ddl_traffic
		self.total_30ddl_traffic_limit = total_30ddl_traffic_limit

	@staticmethod
	def from_dict(data: Dict):
		return Account(token=data["token"], email=data["email"],
		               tier=data["tier"], root_folder=data["rootFolder"],
		               files_count=data["filesCount"], files_count_limit=data["filesCountLimit"],
		               total_size=data["totalSize"], total_size_limit=data["totalSizeLimit"],
		               total_30ddl_traffic=data["total30DDLTraffic"],
		               total_30ddl_traffic_limit=data["total30DDLTrafficLimit"])



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from datetime import datetime
import requests
from sebaubuntu_libs.liblineage import GITHUB_ORG
from typing import Dict, List, Union
import yaml

class BatteryData:
	"""LineageOS battery information.

	Attributes:
	- capacity (int): The battery capacity (mAh)
	- removable (bool): Whether the battery is removable
	"""
	def __init__(self, capacity: int, removable: bool):
		"""Initialize the battery information."""
		self.capacity = capacity
		self.removable = removable

	@classmethod
	def from_dict(self, battery: Dict):
		"""Create a battery information object from a dictionary."""
		return self(battery["capacity"], battery["removable"])

	def __str__(self) -> str:
		"""Return a string representation of the battery information."""
		return (f"capacity: {self.capacity}mAh"
		        f", removable: {self.removable}")

class BluetoothData:
	"""LineageOS Bluetooth information.

	Attributes:
	- spec (str): Bluetooth specification
	"""
	def __init__(self, spec: str):
		"""Initialize the Bluetooth information."""
		self.spec = spec

	@classmethod
	def from_dict(self, bluetooth: Dict):
		"""Create a Bluetooth information object from a dictionary."""
		return self(bluetooth["spec"])

	def __str__(self) -> str:
		return f"spec: {self.spec}"

class DimensionData:
	"""LineageOS dimension information.

	Attributes:
	- height (str): The height
	- width (str): The width
	- depth (str): The depth
	"""
	def __init__(self, height: str, width: str, depth: str):
		"""Initialize the dimension information."""
		self.height = height
		self.width = width
		self.depth = depth

	@classmethod
	def from_dict(self, dimension: Dict):
		"""Create a dimension information object from a dictionary."""
		return self(dimension["height"], dimension["width"], dimension["depth"])

	def __str__(self) -> str:
		return (f"height: {self.height}"
		        f", width: {self.width}"
		        f", depth: {self.depth}")

class ScreenData:
	"""LineageOS screen information.

	Attributes:
	- size (str): The screen size (inches)
	- density (int): The screen density (dpi)
	- resolution (str): The screen resolution (e.g. 1080x1920)
	- technology (str): The screen technology (e.g. LCD)
	"""
	def __init__(self, size: str, density: int, resolution: str, technology: str):
		"""Initialize the screen information."""
		self.size = size
		self.density = density
		self.resolution = resolution
		self.technology = technology

	@classmethod
	def from_dict(self, device: Dict):
		"""Create a screen information object from a dictionary."""
		return self(device["size"], device["density"], device["resolution"], device["technology"])

	def __str__(self) -> str:
		return (f"size: {self.size}"
		        f", density: {self.density}"
		        f", resolution: {self.resolution}"
		        f", technology: {self.technology}")

class DeviceData:
	"""LineageOS wiki device data.

	Attributes:
	- architecture (str): The architecture of the device
	- battery (BatteryData | dict[str, BatteryData] | None): Battery info
	- bluetooth (BluetoothData): Bluetooth support info
	- codename (str): The codename of the device
	- cpu: (str): CPU name
	- cpu_cores (str): Number of CPU cores
	- cpu_freq (str): CPU frequency
	- current_branch (str): The current branch of the device
	- dimensions (DimensionData | dict[str, DimensionData] | None): Dimensions of the device
	- gpu (str): GPU name
	- image (str): The image of the device
	- install_method (str): The install method of the device
	- kernel (str): Kernel repository
	- maintainers (list[str]): The maintainers of the device
	- name (str): Commercial name of the device
	- peripherals (list[str]): Peripherals supported by the device
	- release (datetime | dict[str, datetime]): The release date of the device
	- screen (ScreenData | dict[str, ScreenData] | None): Screen info
	- tree (str): Device tree repository of the device
	- type (str): The form factor of the device
	- vendor (str): Brand name of the device vendor
	- vendor_short (str): Short name of the device vendor
	- versions (list[str]): The versions of the device
	"""
	def __init__(self,
	             architecture: str,
	             battery: Union[BatteryData, Dict, None],
	             bluetooth: BluetoothData, codename: str,
	             cpu: str,
	             cpu_cores: str,
	             cpu_freq: str,
	             current_branch: float,
	             dimensions: Union[DimensionData, Dict, None],
	             gpu: str,
	             image: str,
	             install_method: str,
	             kernel: str,
	             maintainers: List,
	             name: str,
	             peripherals: List,
	             release: Union[datetime, Dict, None],
	             screen: Union[ScreenData, Dict, None],
	             tree: str,
	             device_type: str,
	             vendor: str,
	             vendor_short: str,
	             versions: List[float]
	            ):
		"""Initialize the device information."""
		self.architecture = architecture
		self.battery = battery
		self.bluetooth = bluetooth
		self.codename = codename
		self.cpu = cpu
		self.cpu_cores = cpu_cores
		self.cpu_freq = cpu_freq
		self.current_branch = current_branch
		self.dimensions = dimensions
		self.gpu = gpu
		self.image = image
		self.install_method = install_method
		self.kernel = kernel
		self.maintainers = maintainers
		self.name = name
		self.peripherals = peripherals
		self.release = release
		self.screen = screen
		self.tree = tree
		self.device_type = device_type
		self.vendor = vendor
		self.vendor_short = vendor_short
		self.versions = versions

	@classmethod
	def from_dict(self, data: Dict):
		"""Create a device data object from a dictionary."""
		if data["battery"]:
			if isinstance(data["battery"], list):
				battery = {}
				for bat in data["battery"]:
					device, battery_data = list(bat.items())[0]
					battery[device] = BatteryData.from_dict(battery_data)
			else:
				battery = BatteryData.from_dict(data["battery"])
		else:
			battery = None

		if data["bluetooth"]:
			bluetooth = BluetoothData.from_dict(data["bluetooth"])
		else:
			bluetooth = None
		
		if data["dimensions"]:
			if isinstance(data["dimensions"], list):
				dimensions = {}
				for device, dimension_data in data["dimensions"]:
					dimensions[device] = DimensionData.from_dict(dimension_data)
			elif isinstance(data["dimensions"], dict):
				dimensions = DimensionData.from_dict(data["dimensions"])
			else:
				dimensions = None
		else:
			dimensions = None

		def convert_release_date(date: Union[int, str]):
			date = str(date)
			# YYYY-MM-DD
			try:
				return datetime.fromisoformat(date)
			except ValueError:
				# YYYY-MM
				try:
					return datetime.strptime(date, "%Y-%m")
				except ValueError:
					# YYYY
					return datetime.strptime(date, "%Y")

		if isinstance(data["release"], list):
			release = {}
			for rel in data["release"]:
				device, date = list(rel.items())[0]
				release[device] = convert_release_date(date)
		else:
			release = convert_release_date(data["release"])

		if data["screen"]:
			if isinstance(data["screen"], list):
				screen = {}
				for scr in data["screen"]:
					device, screen_data = list(scr.items())[0]
					screen[device] = ScreenData.from_dict(screen_data)
			else:
				screen = ScreenData.from_dict(data["screen"])
		else:
			screen = None

		return self(
			data["architecture"],
			battery,
			bluetooth,
			data["codename"],
			data["cpu"],
			data["cpu_cores"],
			data["cpu_freq"],
			data["current_branch"],
			dimensions,
			data["gpu"],
			data["image"],
			data["install_method"],
			data["kernel"],
			data["maintainers"],
			data["name"],
			data["peripherals"],
			release,
			screen,
			data["tree"],
			data["type"],
			data["vendor"],
			data["vendor_short"],
			data["versions"],
		)

	def __str__(self) -> str:
		"""Return a string representation of the device data."""
		return (f"Name: {self.name}\n"
		        f"Codename: {self.codename}\n"
		        f"Type: {self.device_type}\n"
				f"Vendor: {self.vendor}\n"
				f"Vendor (short): {self.vendor_short}\n"
				f"Architecture: {self.architecture}\n"
				f"CPU: {self.cpu}\n"
				f"CPU cores: {self.cpu_cores}\n"
				f"CPU frequency: {self.cpu_freq}\n"
				f"GPU: {self.gpu}\n"
				f"Install method: {self.install_method}\n"
				f"Kernel repository: {f'{GITHUB_ORG}/{self.kernel}'}\n"
				f"Maintainers: {', '.join(self.maintainers) if self.maintainers else 'None (unmaintained)'}\n"
				f"Peripherals: {', '.join(self.peripherals)}\n"
				f"Release: {str(self.release) if not isinstance(self.release, dict) else ', '.join([f'{device}: {date}' for device, date in self.release.items()])}\n"
				f"Screen: {str(self.screen) if not isinstance(self.screen, dict) else ', '.join([f'{device}: {screen_data}' for device, screen_data in self.screen.items()])}\n"
				f"Device tree repository: {f'{GITHUB_ORG}/{self.tree}'}\n"
				f"Versions: {', '.join([str(version) for version in self.versions])}\n"
				f"Battery: {str(self.battery) if not isinstance(self.battery, dict) else ', '.join([f'{device}: {battery_data}' for device, battery_data in self.battery.items()])}\n"
				f"Bluetooth: {str(self.bluetooth)}\n"
				f"Dimensions: {str(self.dimensions) if not isinstance(self.dimensions, dict) else ', '.join([f'{device}: {dimensions_data}' for device, dimensions_data in self.dimensions.items()])}")

def get_device_data(device: str):
	url = f"https://raw.githubusercontent.com/LineageOS/lineage_wiki/master/_data/devices/{device}.yml"
	response = requests.get(url=url)
	response.raise_for_status()
	return DeviceData.from_dict(yaml.safe_load(response.text))



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""nekobin library."""

from requests import post

URL = "https://nekobin.com"

def to_nekobin(data: str) -> str:
	"""Upload a string to Nekobin and return its URL."""
	data = {"content": data}

	resp = post(f"{URL}/api/documents", json=data)
	resp.raise_for_status()
	resp_json = resp.json()

	key = resp_json.get("result").get("key")
	return f"{URL}/{key}"



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""Exceptions utils."""

import traceback

def format_exception(exception):
	"""Format an exception."""
	return ''.join(traceback.format_exception(type(exception), exception,
	                                          exception.__traceback__,
	                                          limit=None, chain=True))



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from typing import List
from sebaubuntu_libs.libvintf import INDENTATION
from sebaubuntu_libs.libvintf.common import Hal, cast_to_str_key
from textwrap import indent
from xml.etree.ElementTree import Element

class HidlInterface:
	"""Class representing a HIDL interface."""
	def __init__(self, name: str, version: str, instance: str):
		"""Initialize an object."""
		self.name = name
		self.version = version
		self.instance = instance

	def __str__(self) -> str:
		return f"@{self.version}::{self.name}/{self.instance}"

	def __eq__(self, __o: object) -> bool:
		if isinstance(__o, HidlInterface):
			return (self.name == __o.name
			        and self.version == __o.version
			        and self.instance == __o.instance)
		return False

	def __hash__(self) -> int:
		return hash((self.name, self.version, self.instance))

	@classmethod
	def from_fqname(cls, fqname: str) -> 'HidlInterface':
		"""Create a AIDL HAL from a FQName."""
		temp = fqname.removeprefix("@")
		version, temp = temp.split("::", 1)
		name, instance = temp.split("/", 1)

		return cls(name, version, instance)

	@classmethod
	def from_interface(cls, version: str, interface: Element) -> List['HidlInterface']:
		"""Create a AIDL HAL from an interface."""
		name = interface.findtext("name")

		return [cls(name, version, instance.text) for instance in interface.findall("instance")]

	@classmethod
	def from_interfaces(cls, version: str, interfaces: List[Element]) -> List['HidlInterface']:
		instances = [cls.from_interface(version, interface) for interface in interfaces]

		return [interface for interfaces in instances for interface in interfaces]

class HidlTransport:
	"""Class representing a HIDL transport type."""
	def __init__(self, name: str, passthrough_arch: str = None):
		"""Initialize an object."""
		self.name = name
		self.passthrough_arch = passthrough_arch

	def __str__(self) -> str:
		if self.name == "passthrough":
			return f'<transport arch="{self.passthrough_arch}">{self.name}</transport>'

		return f'<transport>{self.name}</transport>'

	def __eq__(self, __o: object) -> bool:
		if isinstance(__o, HidlTransport):
			return (self.name == __o.name
			        and self.passthrough_arch == __o.passthrough_arch)
		return False
	
	def __hash__(self) -> int:
		return hash((self.name, self.passthrough_arch))

	@classmethod
	def from_element(cls, element: Element):
		"""Get a HidlTransport from an XML element."""
		name = element.text
		passthrough_arch = None
		if name == "passthrough":
			passthrough_arch = element.get("arch")

		return cls(name, passthrough_arch)

class HidlHal(Hal):
	"""Class representing a HIDL HAL."""
	def __init__(self, name: str, transport: HidlTransport, interfaces: set[HidlInterface]):
		"""Initialize an object."""
		super().__init__(name)

		self.transport = transport
		self.interfaces = interfaces

	def __eq__(self, __o: object) -> bool:
		if isinstance(__o, HidlHal):
			return (self.name == __o.name
			        and self.transport == __o.transport
			        and self.interfaces == __o.interfaces)
		return False

	def __hash__(self) -> int:
		return hash((self.name, self.transport, self.interfaces))

	def __str__(self) -> str:
		string = f'<hal format="hidl">\n'
		string += indent(f'<name>{self.name}</name>\n', INDENTATION)
		string += indent(f'{self.transport}\n', INDENTATION)
		for interface in sorted(self.interfaces, key=cast_to_str_key):
			string += indent(f'<fqname>{interface}</fqname>\n', INDENTATION)
		string += '</hal>'

		return string

	@classmethod
	def from_entry(cls, entry: Element) -> 'HidlHal':
		"""Create a HIDL HAL from a VINTF entry."""
		assert entry.get("format") == "hidl"

		name = entry.findtext("name")
		transport = HidlTransport.from_element(entry.find("transport"))

		version = entry.findtext("version")
		interfaces = [
			HidlInterface.from_fqname(interface.text) for interface in entry.findall("fqname")
		]
		interfaces.extend(HidlInterface.from_interfaces(version, entry.findall("interface")))
		interfaces = set(interfaces)

		return cls(name, transport, interfaces)



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""LineageOS library."""

from sebaubuntu_libs.libandroid.versions import AndroidVersion

GITHUB_ORG = "https://github.com/LineageOS"

LINEAGEOS_TO_ANDROID_VERSION = {
    "13.0": AndroidVersion.M,
    "14.1": AndroidVersion.N,
    "15.1": AndroidVersion.O,
	"16.0": AndroidVersion.P,
	"17.1": AndroidVersion.Q,
	"18.0": AndroidVersion.R,
	"18.1": AndroidVersion.R,
	"19.0": AndroidVersion.S,
	"19.1": AndroidVersion.Sv2,
}



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from datetime import datetime
from typing import Dict
import requests

API_VERSION = "v1"
API_URL = f"https://download.lineageos.org/api/{API_VERSION}"

class FullUpdateInfo:
	"""LineageOS full update information.

	Attributes:
	- datetime (datetime): The date and time of the update
	- filename (str): The filename of the update
	- id (str): The ID of the update
	- romtype (str): The ROM type of the update
	- size (int): The size of the update (bytes)
	- url (str): The URL of the update
	- version (str): The LineageOS version of the update (e.g. 18.1)
	"""
	def __init__(self,
		         datetime: datetime,
		         filename: str,
				 id: str,
				 romtype: str,
				 size: int,
				 url: str,
				 version: str,
		        ):
		"""Initialize the full update information."""
		self.datetime = datetime
		self.filename = filename
		self.id = id
		self.romtype = romtype
		self.size = size
		self.url = url
		self.version = version

	@classmethod
	def from_json(self, update: Dict):
		"""Create a full update information object from a JSON object."""
		return self(datetime.fromtimestamp(update["datetime"]), update["filename"],
		            update["id"], update["romtype"],
		            update["size"], update["url"], update["version"])

def get_nightlies(device: str):
	"""Get the latest OTAs for a device."""
	url = f"{API_URL}/{device}/nightly/1"
	response = requests.get(url=url).json()["response"]
	updates = [FullUpdateInfo.from_json(update) for update in response]

	return updates
